'''
Brief:
    File for the NVMe Devices

Author(s):
    Charles Machalow
'''

import collections
import time
from ctypes import *

from pyrw.rwe_parser import *
from pyrw.pci import PCIDevice

COMPLETION_ENTRY_SIZE = 16
SUBMISSION_ENTRY_SIZE = 64
NVME_CLASS_CODE = 0x10802

AdminQueueAddresses = collections.namedtuple("AdminQueueAddresses", ['AdminSubmissionQueueBase', 'AdminCompletionQueueBase'])
AdminQueueAttributes = collections.namedtuple("AdminQueueAttributes", ['AdminCompletionQueueSize', 'AdminSubmissionQueueSize'])
AdminQueueEntries = collections.namedtuple("AdminQueueEntries", ['AdminCompletionQueue', 'AdminSubmissionQueue'])
NVMeVersion = collections.namedtuple("NVMeVersion", ['Major', 'Minor', 'Tertiary'])

def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

class NVMeDevice(PCIDevice):
    '''
    Brief:
        Abstraction for an NVMe PCIe Device
    '''
    def __init__(self, rwe, address):
        '''
        Brief:
            Initializer for the NVMe Device
        '''
        PCIDevice.__init__(self, rwe, address)
        assert self.getPCIClassCode() == NVME_CLASS_CODE, "This is not an NVMe device."

    def getControllerRegisterData(self):
        '''
        Brief:
            Returns a copy of the first 4096 bytes of controller register data
        '''
        bar0 = self.getPCIBarAddresses()[0]
        return self.rwe.readMemory(bar0, 4096)

    def getNVMeVersion(self):
        '''
        Brief:
            Returns the NVMe version from the controller registers
        '''
        controllerRegisterData = self.getControllerRegisterData()
        vs = bytesToDWordList(controllerRegisterData[0x8:])[0]
        return NVMeVersion(Major=vs >> 16, Minor=((vs >> 8) & 0xFF), Tertiary=vs & 0xFF)

    def getAdminQueueAttributes(self):
        '''
        Brief:
            Returns the admin queue sizes
        '''
        controllerRegisterData = self.getControllerRegisterData()
        aqa = bytesToDWordList(controllerRegisterData[0x24:])[0]
        return AdminQueueAttributes(AdminCompletionQueueSize=(aqa >> 16) & 0xFFF, AdminSubmissionQueueSize=aqa & 0xFFF)

    def getAdminQueueBaseAddresses(self):
        '''
        Brief:
            Returns the addresses of the admin submission and completion queues
        '''
        controllerRegisterData = self.getControllerRegisterData()
        asqb = ((c_uint64.from_buffer_copy(controllerRegisterData[0x28:]).value) >> 12) << 12
        acqb = ((c_uint64.from_buffer_copy(controllerRegisterData[0x30:]).value) >> 12) << 12
        return AdminQueueAddresses(asqb, acqb)

    def getAdminQueueEntries(self):
        '''
        Brief:
            Returns all the queue entries in the admin queues
        '''
        addresses = self.getAdminQueueBaseAddresses()
        sizes = self.getAdminQueueAttributes()
        submissionQueue = self.rwe.readMemory(addresses.AdminSubmissionQueueBase, SUBMISSION_ENTRY_SIZE * sizes.AdminSubmissionQueueSize)
        completionQueue = self.rwe.readMemory(addresses.AdminCompletionQueueBase, COMPLETION_ENTRY_SIZE * sizes.AdminCompletionQueueSize)
        sqDw = bytesToDWordList(submissionQueue)
        cqDw = bytesToDWordList(completionQueue)
        return AdminQueueEntries(list(chunks(cqDw, 4)), list(chunks(sqDw, 16)))

    def controllerReset(self):
        '''
        Brief:
            Performs a complete NVMe Controller Reset.
                CC.EN -> 0
                CSTS.RDY -> 0
                CC.EN -> 1
                CSTS.RDY -> 1
        '''
        bar0 = self.getPCIBarAddresses()[0]
        CCAddr = bar0 + 0x14
        CSTSAddr = bar0 + 0x1C

        timeoutMs = self.getControllerRegisterData()[3] * 500

        CC0 = self.rwe.readMemory(CCAddr, 1)[0]
        CC0 = ((CC0 >> 1) << 1) # set CC.EN to 0
        assert self.rwe.writeMemory(CCAddr, [CC0]).ReturnCode == 0

        deathTime = time.time() + (timeoutMs / 1000.0)
        while time.time() < deathTime:
            CSTS0 = self.rwe.readMemory(CSTSAddr, 1)[0]
            if CSTS0 & 1 == 0:
                break
        else:
            raise RuntimeError("CSTS.RDY did not go to 0")

        CC0 = CC0 + 1 # set CC.EN to 1
        self.rwe.writeMemory(CCAddr, [CC0])

        deathTime = time.time() + (timeoutMs / 1000.0)
        while time.time() < deathTime:
            CSTS0 = self.rwe.readMemory(CSTSAddr, 1)[0]
            if CSTS0 & 1 == 1:
                break
        else:
            raise RuntimeError("CSTS.RDY did not go back to 1")

    def subsystemReset(self):
        '''
        Brief:
            Performs an NVMe Subsystem Reset
        '''
        # write 4E564D65h ("NVMe") to NSSR (0x20)
        bar0 = self.getPCIBarAddresses()[0]
        NSSR = bar0 + 0x20

        if self.rwe.writeMemory(NSSR, [0x4e, 0x56, 0x4d, 0x65]).ReturnCode != 0:
            raise RuntimeError("Failed to write the NSSR")

        # todo: check if it worked?


'''
Brief:
    Contains helpers to parse RWE output data

Author(s):
    Charles Machalow
'''
import collections
import re

PCI_DEVICE_LINE_REGEX = r'Bus (\w*), Device (\w*), Function (\w*) \- (.*)'
RWE_ADDRESS_REGEX = r' Address=(\w*?), '
PCILocation = collections.namedtuple("PCILocation", ['Bus', 'Device', 'Function'])

def getBinaryFromHexDump(hexDumpStr):
    '''
    Brief:
        Gets a binary representation of the string hex dump RWE would print in some cases
    '''
    lines = hexDumpStr.splitlines()
    lines = [line for line in lines if line[0] != ' '] # remove spaced lines
    lines = [line for line in lines if line[1] != 'u'] # remove dump lines
    lines = [line for line in lines if line[2] != 'r'] # remove parameter error lines
    ret = []
    for line in lines:
        line = line.split(' ', 1)[1] # remove index
        line = line.split('\t')[0]   # remove ascii
        ret += [int(c, 16) for c in line.split()]

    return bytearray(ret)

def bytesToDWordList(b):
    '''
    Brief:
        Convert a list of bytes to a list of DWords
    '''
    ret = []
    for idx in range(0, len(b), 4):
        d = b[idx] + (b[idx + 1] << 8) + (b[idx + 2] << 16) + (b[idx + 3] << 24)
        ret.append(d)

    return ret

def pciTreeTextToDict(txt):
    '''
    Brief:
        Converts a PCI tree's text to a dict of PCI devices to descriptions
    '''
    a = re.findall(PCI_DEVICE_LINE_REGEX, txt)
    retDict = {}
    for itm in a:
        retDict[
            PCILocation(int(itm[0], 16), int(itm[1], 16), int(itm[2], 16))
        ] = itm[3].strip()

    return retDict

def verifyAddress(address, rweOutput):
    '''
    Brief:
        Verifies that the given address matches the RWE output. Raises if it is mismatched.
            Note that there are some current issues with 64-bit addresses and RWE not handling them properly in commands.
                This should at least tell the user if that happens, though there is no current workaround.
                    I've emailed RWE's Jeff about this bug. Said it should be fixed in a future release.
    '''
    m = re.findall(RWE_ADDRESS_REGEX, rweOutput)
    rweAddr = int(m[0], 16)
    if address != rweAddr:
        raise RuntimeError("RWE Didn't process the correct address! It processed: 0x%X instead of 0x%X" % (rweAddr, address))




from pyrw.rwe import ReadWriteEverything

rwe = ReadWriteEverything()
header='''
| --------------------------------------------- |
| rwe object has been created for your usage... |
| %-45s |
| --------------------------------------------- |''' % rwe.getRWEVersion()
print(header)

try:
    import IPython
    start_interactive = lambda: IPython.embed(display_banner=False)
except ImportError:
    print("| Install IPython to have a more complete       |")
    print("|     interactive experience.                   |")
    print("| --------------------------------------------- |")
    import code
    start_interactive = lambda: code.interact(banner='', local=dict(rwe=rwe))

start_interactive()



'''
Brief:
    File for the PCI device

Author(s):
    Charles Machalow
'''

class PCIDevice(object):
    '''
    Brief:
        Abstraction of some things to do on a PCI device
    '''
    def __init__(self, rwe, address):
        '''
        Brief:
            Initializer for the object. Takes a RWE instance and the PCI Address (PCILocation)
        '''
        self.rwe = rwe
        self.address = address

        # aliases
        self.readPCI = lambda : self.rwe.readPCI(self.address.Bus, self.address.Device, self.address.Function)
        self.writePCI = lambda data: rwe.writePCI(self.address.Bus, self.address.Device, self.address.Function, data)
        self.getPCIClassCode = lambda : self.rwe.getPCIClassCode(self.address.Bus, self.address.Device, self.address.Function)
        self.getPCIBarAddresses = lambda : self.rwe.getPCIBarAddresses(self.address.Bus, self.address.Device, self.address.Function)


'''
Brief:
    finder.py - Contains utilities related to finding the RW-Everything executable

Author(s):
    Charles Machalow
'''

import os

from pyrw.rwe import ReadWriteEverything

THIS_FOLDER = os.path.abspath(os.path.dirname(__file__))
BIN_FOLDER = os.path.join(THIS_FOLDER, 'bin')
X86_BIN_FOLDER = os.path.join(BIN_FOLDER, 'x86')
X64_BIN_FOLDER = os.path.join(BIN_FOLDER, 'x64')

X64_PROGRAM_FILES = '/Program Files (x86)'

RW_EXE = 'Rw.exe'
DEFAULT_INSTALL_LOCATIONS = [
    "/Program Files/RW-Everything/%s" % RW_EXE,
    "/Program Files (x86)/RW-Everything/%s" % RW_EXE,
]

def findInstalledRWEverything():
    '''
    Brief:
        Searches for an installed version of RW-Everything. If found, returns the string path.
    '''
    drivePath = os.environ.get('SYSTEMDRIVE', 'C:')
    for defInstLoc in DEFAULT_INSTALL_LOCATIONS:
        fullPath = os.path.join(drivePath, defInstLoc)
        if os.path.isfile(fullPath):
            try:
                ReadWriteEverything(fullPath)
                return fullPath
            except EnvironmentError:
                continue

def getEnvironmentVariableAsList(v):
    '''
    Brief:
        Returns the given environment variable's values as a list
    '''
    return os.environ.get(v, '').split(os.pathsep)

def findPathedRWEverything():
    '''
    Brief:
        Searched for a RW-Everything executable in the system path. If found, returns the string path.
    '''
    path = getEnvironmentVariableAsList('PATH')
    for p in path:
        possibleRw = os.path.join(p, RW_EXE)
        if os.path.isfile(possibleRw):
            try:
                ReadWriteEverything(possibleRw)
                return possibleRw
            except EnvironmentError:
                continue

def findCwdRWEverything():
    '''
    Brief:
        Searches the current working directory for RWE
    '''
    fullPath = os.path.join(os.getcwd(), RW_EXE)
    if os.path.isfile(fullPath):
        try:
            ReadWriteEverything(fullPath)
            return fullPath
        except EnvironmentError:
            pass

def findPackagedRWEverything():
    '''
    Brief:
        Finds the RWE packaged with this module
    '''
    folder = X86_BIN_FOLDER
    if os.environ.get('ProgramFiles(x86)', False):
        # 64 bit
        folder = X64_BIN_FOLDER

    loc = os.path.join(folder, RW_EXE)
    if os.path.isfile(loc):
        return loc

def findRWEverything():
    '''
    Brief:
        Searches known directories for RWE
    '''
    r = findCwdRWEverything()
    if r:
        return r
    
    r = findPathedRWEverything()
    if r:
        return r

    r = findPackagedRWEverything()
    if r:
        return r

    r = findInstalledRWEverything()

    return r




'''
Brief:
    File for the ReadWriteEverything class

Author(s):
    Charles Machalow
'''
import collections
import logging
import os
import subprocess
import tempfile

from ctypes import *

from pyrw.rwe_parser import bytesToDWordList, pciTreeTextToDict, verifyAddress
from pyrw.nvme import NVMeDevice, NVME_CLASS_CODE

logger = logging.getLogger(__name__)

THIS_FOLDER = os.path.abspath(os.path.dirname(__file__))
LOCAL_TMP_FILE = os.path.join(THIS_FOLDER, '__tmp.bin')
ProcessOutput = collections.namedtuple("ProcessOutput", ["Output", "ReturnCode"])

class ReadWriteEverything(object):
    '''
    Brief:
        Easy to use abstractions for RWE in Python
    '''
    def __init__(self, exePath=None):
        '''
        Brief:
            Initializer for the class. Takes the path to rw.exe
                Ensures we have admin and that rw.exe seems to work
        '''

        if exePath is None:
            # import here to prevent a circle
            from pyrw.finder import findRWEverything
            exePath = findRWEverything()

        self.exePath = exePath
        self.version = self.getRWEVersion()

        if not windll.shell32.IsUserAnAdmin():
            raise EnvironmentError("Please run as admin")

        self._checkValidRWExe()

    def _checkValidRWExe(self):
        '''
        Brief:
            Does a stupidly silly check to see if rw.exe appears to work
        '''
        r = self.callRWECommand("COUT Hello World;rwexit")
        if 'RW Exit' in r.Output and 'Hello World' in r.Output and r.ReturnCode:
            raise EnvironmentError("%s does not appear to be a valid RW-Everything exe" % self.exePath)

    def getRWEVersion(self):
        '''
        Brief:
            Uses some magic to get the RWE version
        '''
        if hasattr(self, 'version'):
            return self.version

        version = subprocess.check_output('powershell "(Get-Item -path \\"%s\\").VersionInfo.FileVersion"' % self.exePath).strip().decode()
        with open(self.exePath, 'rb') as f:
            # https://superuser.com/questions/358434/how-to-check-if-a-binary-is-32-or-64-bit-on-windows
            f.seek(0x204)
            if f.read(1)[0] == 0x64:
                b = 'x64'
            else:
                b = 'x86'

        v = 'RW - Read Write Utility v%s %s' % (version, b)
        self.version = v
        logger.debug("RWE Version: %s" % v)
        return self.version

    def callRawCommand(self, cmd):
        '''
        Brief:
            Calls a Raw command on rw.exe
        '''
        fullCmd = '\"%s\" %s' % (self.exePath, cmd)
        logger.debug("Calling raw command: %s" % fullCmd)
        try:
            output = subprocess.check_output(fullCmd, shell=True, stderr=subprocess.STDOUT)
            retCode = 0
        except subprocess.CalledProcessError as ex:
            output = ex.output
            retCode = ex.returncode

        ret = ProcessOutput(Output=output.decode(), ReturnCode=retCode)
        logger.debug("... Returned: %s" % str(ret))
        return ret

    def callRWECommand(self, cmd):
        '''
        Brief:
            Calls an embeded RWE command on rw.exe
        '''
        fullCommand = '/Min /Nologo /Stdout /Command="%s"' % (cmd.replace('\"', '\\"'))
        return self.callRawCommand(fullCommand)

    def readMemory(self, byteOffset, numBytes):
        '''
        Brief:
            Reads raw memory from a given offset for a given number of bytes
        '''
        n = self.callRWECommand("SAVE \"%s\" Memory 0x%X %d" % (LOCAL_TMP_FILE, byteOffset, numBytes))
        assert n.ReturnCode == 0, "Didn't return 0"
        verifyAddress(byteOffset, n.Output)

        with open(LOCAL_TMP_FILE, 'rb') as f:
            data = f.read()
        os.remove(LOCAL_TMP_FILE)
        return data

    def writeMemory(self, byteOffset, data):
        '''
        Brief:
            Writes given data to the given offset of memory
        '''
        with open(LOCAL_TMP_FILE, 'wb') as f:
            f.write(bytearray(data))

        ret = self.callRWECommand('LOAD "%s" Memory %d' % (LOCAL_TMP_FILE, byteOffset))
        verifyAddress(byteOffset, ret.Output)

        os.remove(LOCAL_TMP_FILE)
        return ret

    def readPCI(self, bus, device, function):
        '''
        Brief:
            Reads PCI header (configuration space) data from the given device
        '''
        n = self.callRWECommand("SAVE \"%s\" PCI %d %d %d" % (LOCAL_TMP_FILE, bus, device, function))
        assert n.ReturnCode == 0, "Didn't return 0"
        with open(LOCAL_TMP_FILE, 'rb') as f:
            data = f.read()
        os.remove(LOCAL_TMP_FILE)
        return data

    def writePCI(self, bus, device, function, data):
        '''
        Brief:
            Writes given PCI (configuration space) data to a given device
        '''
        with open(LOCAL_TMP_FILE, 'wb') as f:
            f.write(bytearray(data))

        ret = self.callRWECommand('LOAD "%s" Memory %d %d %d' % (LOCAL_TMP_FILE, bus, device, function))
        os.remove(LOCAL_TMP_FILE)
        return ret

    def getPCIBarAddresses(self, bus, device, function):
        '''
        Brief:
            Gets a list of BAR addresses for the given device
        '''
        pciRegisters = self.readPCI(bus, device, function)
        pciRegistersAsDWords = bytesToDWordList(pciRegisters[0x10:])
        return [((x >> 4) << 4) for x in pciRegistersAsDWords[:6]] # 6 bars max

    def getPCITree(self):
        '''
        Brief:
            Parses the system's PCI devices to a dictionary
        '''
        n = self.callRWECommand('PCITREE')
        assert n.ReturnCode == 0, "Didn't return 0"
        return pciTreeTextToDict(n.Output)

    def getPCIClassCode(self, bus, device, function):
        '''
        Brief:
            Gets the PCI class code for a given device
        '''
        d = self.readPCI(bus, device, function)
        return bytesToDWordList(d[0x09:0x0D])[0] & 0xFFFFFF

    def getNVMeDevices(self):
        '''
        Brief:
            Returns a list of all NVMe devices on the system
        '''
        nvmeDevices = []
        for address in self.getPCITree().keys():
            testClassCode = self.getPCIClassCode(address.Bus, address.Device, address.Function)
            if testClassCode == NVME_CLASS_CODE:
                nvmeDevices.append(NVMeDevice(self, address))

        return nvmeDevices

if __name__ == '__main__':
    rwe = ReadWriteEverything()


from setuptools import setup

setup(
    name='pyrw',
    author='csm10495',
    author_email='csm10495@gmail.com',
    url='http://github.com/csm10495/pyrw',
    version='0.8',
    packages=['pyrw'],
    license='MIT License',
    python_requires='>=2.7',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    include_package_data = True,
)